from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Literal, Mapping

from pydantic import BaseModel

HttpMethod = Literal["GET", "POST", "PUT", "PATCH", "DELETE", "HEAD", "OPTIONS"]

def _clean_dict(data: Mapping[str, Any]) -> dict[str, Any]:
    return {key: value for key, value in data.items() if value is not None}


def _require_non_empty(value: str, field: str) -> None:
    if not value or not str(value).strip():
        raise ValueError(f"{field} is required.")

def _normalize_path(path: str) -> str:
    normalized = "/" + path.lstrip("/")
    if normalized == "/":
        raise ValueError("Endpoint path cannot be empty.")
    return normalized


def _join_base_path(base_path: str, endpoint_path: str) -> str:
    base = base_path.strip("/")
    tail = endpoint_path.lstrip("/")
    if not base:
        return "/" + tail
    return f"/{base}/{tail}"

class Endpoint(BaseModel):
    """
    Object representing an API endpoint

    :arg name: Name of the endpoint
    :type name: str
    :arg method: HTTP method for the endpoint
    :type method: HttpMethod
    :arg path: Endpoint path
    :type path: str
    :raises pydantic.ValidationError: If http method is invalid
    :raises ValueError: If endpoint path is empty
    """
    name: str
    method: HttpMethod
    path: str

    def __post_init__(self) -> None:
        object.__setattr__(self, "path", _normalize_path(self.path))

    def full_path(self, base_path: str = "") -> str:
        """
        Returns the full path for the endpoint

        :param base_path: Base path for the endpoint, defaults to an empty string
        :type base_path: str
        :return: Full path for the endpoint
        :rtype: str
        """
        return _join_base_path(base_path, self.path)



@dataclass(slots=True)
class TicketCreateTicket:
    """
    Represents the metadata for a ticket

    :param Title: Title of the ticket
    :type Title: str
    :param Queue: Queue of the ticket
    :type Queue: str
    :param State: State of the ticket
    :type State: str
    :param Priority: Priority of the ticket
    :type Priority: str
    :param CustomerUser: Customer user of the ticket
    :type CustomerUser: str | None
    :param Type: Type of the ticket
    :type Type: str | None
    :param Service: Service of the ticket
    :type Service: str | None
    :param SLA: SLA of the ticket
    :type SLA: str | None
    :param Owner: Owner of the ticket
    :type Owner: str | None
    :raises ValueError: If any required field is empty
    """
    Title: str
    Queue: str
    State: str
    Priority: str
    CustomerUser: str | None = None
    Type: str | None = None
    Service: str | None = None
    SLA: str | None = None
    Owner: str | None = None
    Responsible: str | None = None

    def validate(self) -> None:
        _require_non_empty(self.Title, "Ticket.Title")
        _require_non_empty(self.Queue, "Ticket.Queue")
        _require_non_empty(self.State, "Ticket.State")
        _require_non_empty(self.Priority, "Ticket.Priority")

    def to_dict(self) -> dict[str, Any]:
        self.validate()
        return _clean_dict(
            {
                "Title": self.Title,
                "Queue": self.Queue,
                "State": self.State,
                "Priority": self.Priority,
                "CustomerUser": self.CustomerUser,
                "Type": self.Type,
                "Service": self.Service,
                "SLA": self.SLA,
                "Owner": self.Owner,
                "Responsible": self.Responsible,
            }
        )


@dataclass(slots=True)
class TicketCreateArticle:
    """
    Represents the article content for a ticket

    :param Subject: Subject of the article
    :type Subject: str
    :param Body: Body of the article
    :type Body: str
    :param ContentType: Content type of the article
    :type ContentType: str
    :param Charset: Charset of the article, defaults to None
    :type Charset: str | None
    :param MimeType: MIME type of the article, defaults to None
    :type MimeType: str | None
    :param SenderType: Sender type of the article, defaults to None
    :type SenderType: str | None
    :param From_: From address of the article, defaults to None
    :type From_: str | None
    :raises ValueError: If any required field is empty
    """
    Subject: str
    Body: str
    ContentType: str
    Charset: str | None = None
    MimeType: str | None = None
    SenderType: str | None = None
    From_: str | None = None

    def validate(self) -> None:
        _require_non_empty(self.Subject, "Article.Subject")
        _require_non_empty(self.Body, "Article.Body")
        _require_non_empty(self.ContentType, "Article.ContentType")

    def to_dict(self) -> dict[str, Any]:
        self.validate()
        return _clean_dict(
            {
                "Subject": self.Subject,
                "Body": self.Body,
                "ContentType": self.ContentType,
                "Charset": self.Charset,
                "MimeType": self.MimeType,
                "SenderType": self.SenderType,
                "From": self.From_,
            }
        )


@dataclass(slots=True)
class TicketCreatePayload:
    """
    Represents the payload for creating a ticket

    :param Ticket: Ticket metadata
    :type Ticket: TicketCreateTicket
    :param Article: Article content
    :type Article: TicketCreateArticle
    :param DynamicField: Dynamic fields for the ticket, defaults to None
    :type DynamicField: Mapping[str, Any] | None
    :param Attachment: Attachments for the ticket, defaults to None
    :type Attachment: list[Mapping[str, Any]] | None
    :param TimeUnit: Time unit for the ticket, defaults to None
    :type TimeUnit: int | None
    """
    Ticket: TicketCreateTicket
    Article: TicketCreateArticle
    DynamicField: Mapping[str, Any] | None = None
    Attachment: list[Mapping[str, Any]] | None = None
    TimeUnit: int | None = None

    def to_dict(self) -> dict[str, Any]:
        return _clean_dict(
            {
                "Ticket": self.Ticket.to_dict(),
                "Article": self.Article.to_dict(),
                "DynamicField": self.DynamicField,
                "Attachment": self.Attachment,
                "TimeUnit": self.TimeUnit,
            }
        )



class TicketUpdateTicket(TicketCreateTicket):
    def validate(self):
        pass


from __future__ import annotations

from collections.abc import Mapping
from typing import Any

import httpx

from .endpoints import (
    _DEFAULT_ENDPOINT_IDENTIFIERS,
    _DEFAULT_ENDPOINTS,
    Endpoint,
    EndpointSetter,
    EndpointsRegistry,
)
from .exceptions import TicketClientError
from .models import HttpMethod
from .routes import SessionRoutes, TicketRoutes


class TicketClient:
    """
    Represents a client for interacting with the Ticket API

    :param base_url: Base URL for the Ticket API which the client will connect to
    :type base_url: str | None
    :param username: Username for authentication
    :type username: str | None
    :param password: Password for authentication
    :type password: str | None
    :param endpoints: Optional custom endpoints registry
    :type endpoints: EndpointsRegistry | None
    :param timeout: Request timeout
    :type timeout: float | None
    :param headers: Optional custom headers
    :type headers: Mapping[str, str] | None
    """
    def __init__(
        self,
        base_url: str | None = None,
        *,
        username: str | None = None,
        password: str | None = None,
        endpoints: EndpointsRegistry | None = None,
        timeout: float | None = None,
        headers: Mapping[str, str] | None = None,
        client: httpx.Client | None = None,
    ) -> None:
        self._endpoints = endpoints or EndpointsRegistry()
        self._endpoint_identifiers = dict(_DEFAULT_ENDPOINT_IDENTIFIERS)
        if client is not None:
            self._client = client
        else:
            client_kwargs: dict[str, Any] = {"timeout": timeout, "headers": headers}
            if base_url is not None:
                client_kwargs["base_url"] = base_url
            self._client = httpx.Client(**client_kwargs)

        self._register_default_endpoints()
        self.ticket = TicketRoutes(self)
        self.session = SessionRoutes(self)
        self.set_endpoint = EndpointSetter(self)
        self.session_id: str | None = None

        if username is not None and password is not None:
            self.login(username, password)

    @property
    def endpoints(self) -> EndpointsRegistry:
        return self._endpoints

    def register_endpoint(self, name: str, method: HttpMethod, path: str) -> Endpoint:
        """
        Registers a custom endpoint for the Ticket API

        :param name: Name of the endpoint
        :type name: str
        :param method: HTTP method for the endpoint
        :type method: HttpMethod
        :param path: Endpoint path
        :type path: str
        :return: Registered endpoint
        :rtype: Endpoint
        """
        return self._endpoints.register(Endpoint(name=name, method=method, path=path))

    def set_endpoint_identifier(self, name: str, identifier: str) -> None:
        """
        Sets a custom identifier for an endpoint

        :param name: Name of the endpoint
        :type name: str
        :param identifier: Custom identifier for the endpoint
        :type identifier: str
        """
        self._endpoint_identifiers[name] = identifier

    def endpoint_identifier(self, name: str) -> str:
        """
        Returns the custom identifier for an endpoint

        :param name: Name of the endpoint
        :type name: str
        :return: Custom identifier for the endpoint
        :rtype: str
        """
        try:
            return self._endpoint_identifiers[name]
        except KeyError as exc:
            raise KeyError(f"Endpoint identifier not registered: {name}") from exc

    def login(self, username: str, password: str) -> httpx.Response:
        """
        Creates a new session with the given username and password.

        :param username: Username for authentication
        :type username: str
        :param password: Password for authentication
        :type password: str
        :return: Response object containing session details
        :rtype: Response
        """
        response = self.session.create(username, password)
        self.session_id = response.json().get("SessionID")
        return response


    def request(
        self,
        endpoint_name: str,
        *,
        method: HttpMethod | None = None,
        path: str | None = None,
        path_params: Mapping[str, Any] | None = None,
        **kwargs: Any,
    ) -> httpx.Response:
        """
        Sends a request to the Ticket API

        :param endpoint_name: Name of the endpoint
        :type endpoint_name: str
        :param method: HTTP method for the request, defaults to None
        :type method: HttpMethod | None
        :param path: Custom endpoint path, defaults to None
        :type path: str | None
        :param path_params: Parameters for the endpoint path, defaults to None
        :type path_params: Mapping[str, Any] | None
        :param kwargs: Additional keyword arguments for the request
        :type kwargs: Any
        :return: Response object
        :rtype: httpx.Response
        """
        endpoint_method = method or self._endpoints.method_for(endpoint_name)
        endpoint_path = path or self._endpoints.path_for(endpoint_name)
        
        
        
        if path_params:
            endpoint_path = endpoint_path.format(**path_params)
        response = self._client.request(endpoint_method, endpoint_path, **kwargs)
        
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError:
            try:
                error = response.json().get("Error")
            except Exception:
                error = response.text
            # TODO: improve error to handle status codes
            self._raise_error(error)
        
        if error := response.json().get("Error"):
            self._raise_error(error)
        
        return response

    def _raise_error(self, error: Mapping[str, Any]) -> None:
        raise TicketClientError(error)

    def close(self) -> None:
        """
        Closes the client connection
        """
        self._client.close()

    def __enter__(self) -> "TicketClient":
        return self

    def __exit__(self, exc_type, exc, traceback) -> None:
        self.close()

    def _register_default_endpoints(self) -> None:
        for name, (method, path) in _DEFAULT_ENDPOINTS.items():
            if not self._endpoints.has(name):
                self._endpoints.register(
                    Endpoint(name=name, method=method, path=path)
                )


from typing import Any, Mapping


class TicketClientError(Exception):
    """
    Error returned by the TicketClient API.
    """
    def __init__(self, error: Mapping[str, Any] | str) -> None:
        
        self.error = error
        code = error.get("ErrorCode") if isinstance(error, dict) else error
        message = error.get("ErrorMessage") or "Unknown error" \
            if isinstance(error, dict) else str(error)
            
        super().__init__(f"{code}: {message}" if code else str(message))


from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Mapping, MutableMapping

from .models import Endpoint, HttpMethod

if TYPE_CHECKING:
    from pyznuny.ticket.client import TicketClient

_DEFAULT_ENDPOINTS = {
    "ticket_create": ("POST", "/Ticket"),
    "ticket_update": ("PATCH", "/Ticket/{ticket_id}"),
    "ticket_get": ("GET", "/Ticket/{ticket_id}"),
    "session_create": ("POST", "/Session"),
}

_DEFAULT_ENDPOINT_IDENTIFIERS = {
    "ticket_update": "ticket_id",
    "ticket_get": "ticket_id",
}



class EndpointSetter:
    """
    Custom endpoint setter for the Ticket API

    :arg client: Ticket client instance
    :type client: TicketClient
    """
    def __init__(self, client: "TicketClient") -> None:
        self._client = client

    def ticket_create(self, *, endpoint: str, method: HttpMethod = "POST") -> Endpoint:
        """
        Sets a custom endpoint for creating tickets

        :param endpoint: Custom endpoint for creating tickets
        :type endpoint: str
        :param method: HTTP method for creating tickets, defaults to POST
        :type method: HttpMethod
        :return: Endpoint object
        :rtype: Endpoint
        """
        return self._client.register_endpoint("ticket_create", method, endpoint)

    def ticket_get(
        self,
        *,
        endpoint: str,
        identifier: str = "ticket_id",
        method: HttpMethod = "GET",
    ) -> Endpoint:
        """
        Sets a custom endpoint for retrieving a ticket.

        :param endpoint: Custom endpoint path
        :type endpoint: str
        :param identifier: Identifier for the ticket ID in the endpoint path
        :type identifier: str
        :param method: HTTP method for the endpoint, defaults to GET
        :type method: HttpMethod
        :return: Registered endpoint
        :rtype: Endpoint
        """
        endpoint_obj = self._client.register_endpoint(
            "ticket_get",
            method,
            endpoint,
        )
        self._client.set_endpoint_identifier("ticket_get", identifier)
        return endpoint_obj
    
    def ticket_update(
        self,
        *,
        endpoint: str,
        identifier: str = "ticket_id",
        method: HttpMethod = "POST",
    ) -> Endpoint:
        """
        Sets a custom endpoint for updating a ticket.

        :param endpoint: Custom endpoint path
        :type endpoint: str
        :param identifier: Identifier for the ticket ID in the endpoint path
        :type identifier: str
        :param method: HTTP method for the endpoint, defaults to POST
        :type method: HttpMethod
        :return: Registered endpoint
        :rtype: Endpoint
        """
        endpoint_obj = self._client.register_endpoint(
            "ticket_update",
            method,
            endpoint,
        )
        self._client.set_endpoint_identifier("ticket_update", identifier)
        return endpoint_obj




class EndpointsRegistry:
    """
    Object representing the registry of API endpoints for the Ticket API

    :arg base_path: Base path for the endpoints, defaults to an empty string
    :type base_path: str
    """
    def __init__(
        self,
        *,
        base_path: str = "",
        endpoints: Iterable[Endpoint] | None = None,
    ) -> None:
        self._base_path = base_path
        self._endpoints: MutableMapping[str, Endpoint] = {}
        if endpoints:
            for endpoint in endpoints:
                self.register(endpoint)

    @property
    def base_path(self) -> str:
        return self._base_path

    @base_path.setter
    def base_path(self, value: str) -> None:
        self._base_path = value

    def register(self, endpoint: Endpoint) -> Endpoint:
        self._endpoints[endpoint.name] = endpoint
        return endpoint

    def configure(self, mapping: Mapping[str, tuple[HttpMethod, str]]) -> None:
        for name, (method, path) in mapping.items():
            self.register(Endpoint(name=name, method=method, path=path))

    def get(self, name: str) -> Endpoint:
        try:
            return self._endpoints[name]
        except KeyError as exc:
            raise KeyError(f"Endpoint not registered: {name}") from exc

    def has(self, name: str) -> bool:
        return name in self._endpoints

    def path_for(self, name: str) -> str:
        endpoint = self.get(name)
        return endpoint.full_path(self._base_path)

    def method_for(self, name: str) -> HttpMethod:
        return self.get(name).method



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping

import httpx

from .models import TicketCreatePayload

if TYPE_CHECKING:
    from .client import TicketClient


class SessionRoutes:
    """
    Object representing the routes for creating a session with the Ticket API, this is
    abstracted by the TicketClient
    """
    def __init__(self, client: "TicketClient") -> None:
        self._client = client

    def create(self, username: str, password: str) -> httpx.Response:
        """
        Creates a new session with the given username and password.
        
        :param username: username for authentication
        :type username: str
        :param password: password for authentication
        :type password: str
        :return: Response object containing session details
        :rtype: Response
        """
        return self._client.request(
            "session_create",
            json={"UserLogin": username, "Password": password},
        )


class TicketRoutes:
    """
    Object representing the routes for the Ticket API, this is abstracted
    by the TicketClient
    """
    def __init__(self, client: "TicketClient") -> None:
        self._client = client

    def create(
        self,
        payload: TicketCreatePayload | Mapping[str, Any] | None = None,
        **payload_kwargs: Any,
    ) -> httpx.Response:
        """
        Creates a new ticket

        :param payload: Ticket creation payload
        :type payload: TicketCreatePayload | Mapping[str, Any] | None
        :param payload_kwargs: Additional keyword arguments for the payload
        :type payload_kwargs: Any
        :return: Response object
        :rtype: httpx.Response
        """
        if payload is None:
            payload_dict = dict(payload_kwargs)
            
        elif isinstance(payload, TicketCreatePayload):
            payload_dict = payload.to_dict()
            payload_dict.update(payload_kwargs)
        else:
            payload_dict = dict(payload)
            payload_dict.update(payload_kwargs)

        payload_dict.update({"SessionID": self._client.session_id})
        return self._client.request("ticket_create", json=payload_dict)
    

    def update(self, ticket_id: str | int , **payload: dict) -> httpx.Response:
        """
        Updates an existing ticket

        :param ticket_id: ID of the ticket to update
        :type ticket_id: str | int
        :param payload: Map of the fields to update
        :type payload: dict
        :return: Response object
        :rtype: httpx.Response
        """
        identifier = self._client.endpoint_identifier("ticket_update")
        
        payload.update({"SessionID": self._client.session_id})
        return self._client.request(
            "ticket_update",
            path_params={identifier: ticket_id},
            json=payload,
        )
        
    def get(self, ticket_id: str | int, 
            dynamic_fields:int=0,
            all_articles:int=0) -> httpx.Response:
        """
        Retrieves an existing ticket

        :param ticket_id: ID of the ticket to retrieve
        :type ticket_id: str | int
        :param dynamic_fields: Number of dynamic fields to retrieve, defaults to 0
        :type dynamic_fields: int
        :param all_articles: Whether to retrieve all articles, 1 for true 0 for false,
        defaults to 0
        :type all_articles: int
        :return: Response object
        :rtype: httpx.Response
        """
        identifier = self._client.endpoint_identifier("ticket_get")
        return self._client.request(
            "ticket_get",
            path_params={identifier: ticket_id},
            params={"SessionID": self._client.session_id,
                    "DynamicFields": dynamic_fields,
                    "AllArticles": all_articles},
        )
